# -*- coding: utf-8 -*-
HELP = """merge ranking

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1wMTlaj4PPBlpkh4VE0A3X7D0ZULWnok8
!pip install ranky

!pip install git+https://github.com/djcunningham0/rankaggregation.git
!pip install mc4


"""
import random, pandas as pd, numpy as np, import scipy, math


try :
   import ranky as rk
   from mc4.algorithm import mc4_aggregator
   import rankaggregation as ra
except :
   pass 


##################################################################################################
def log(*s): 
  print(*s, flush=True)


def test():
    # evaluate each rank aggregation algorithm by using spearman's rho's and kendall-tau's metrics
    # for varying levels of ncorrect in the generated rankings
    pd.set_option('display.max_columns', 7)
    for ncorrect in [0, 20, 50, 80, 100]:
      print('ncorrect = {}'.format(ncorrect))

      #### Fake with ncorrect 
      df, rank1, rank2, dict_full = rank_generatefake(ncorrect, 100)
      rank_true = list(dict_full)

      #### Algo merged
      borda     = rank_merge(df, method='borda')
      dowdall   = rank_merge(df, method='dowdall')
      mc4_rank  = rank_merge(df, method='mc4')
      custom    = rank_merge_v2(rank1, rank2, 100)
      avg       = rank_merge(df, method='average')
      df2 = pd.DataFrame([borda, dowdall, avg, mc4_rank, custom, rank1, rank2],
                         index = ['borda', 'dowdall', 'average', 'MC4', 'algo v2', 'rank1', 'rank2'])
      df2 = df2.transpose()

      #### Fill NA with Max ranking as default 
      for idx, col in enumerate(df2.columns):
        max_value = df2[col].max()
        df2[col]  = df2[col].replace(np.nan, max_value)


      dfres = rank_eval(rank_true, df2, 100)
      log(dfres)



############################################################################################
############################################################################################
def rank_generate_fake(dict_full, list_overlap, nsize=100, ncorrect=20):
    """  Returns a list of random rankings of size nsize where ncorrect
         elements have correct ranks

        Keyword arguments:
        dict_full: a dictionary of 1000 objects and their ranks
        list_overlap: list items common to all lists
        nsize: the total number of elements to be ranked
        ncorrect: the number of correctly ranked objects 
    """
    # first randomly sample nsize - len(list_overlap) elements from dict_full
    # of those, ncorrect of them must be correctly ranked
    random_vals = []
    while len(random_vals) <= nsize - len(list_overlap):
      rand = random.sample(list(dict_full), 1)
      if (rand not in random_vals):
        random_vals.append(rand[0])

    # next create list as aggregate of random_vals and list_overlap
    list2 = random_vals + list_overlap
    
    # shuffle nsize - ncorrect elements from list2 
    copy = list2[0:nsize - ncorrect]
    random.shuffle(copy)
    list2[0:nsize - ncorrect] = copy

    # ensure there are ncorrect elements in correct places
    if ncorrect == 0: 
      return list2
    rands = random.sample(list(dict_full)[0:nsize + 1], ncorrect + 1)
    for r in rands:
      list2[r] = list(dict_full)[r]
    return list2



def rank_fillna(df):
    """Replace NaN value with maximum value of column
        Keyword arguments:
        df: a dataframe where each row represents an item and each column represents it's
            assigned rank by a ranker
    """
    cols = df.columns
    for col in cols:
      max_val = df[col].max()
      df[col] = df[col].replace(np.nan, max_val)
    return df



def rank_generatefake(ncorrect, nrank):
    """
      Generate a fake rank list of size nrank that contains ncorrect elements that
      are correctly ranked. Returns a dataframe where the two fake ranks are merged

      # each column of the dataframe should contain the rank each ranker assigns to each candidate
      # because of the fake generate algorithm, some items might not have ranks assigned to them by the ranker

      Keyword arguments:
      ncorrect: The number of correctly ranked elements in list
      nrank: Total number of elements to be ranked
    """
    dict_full = {}
    for i in range(1000):
      dict_full[i] = i + 1

    list_overlap = [1, 15, 20, 30, 80]
    list1 = rank_generate_fake(dict_full, list_overlap, ncorrect=ncorrect) 
    list2 = rank_generate_fake(dict_full, list_overlap, ncorrect=ncorrect)
    rank1, rank2 = [None] * 1000, [None] * 1000
    for i in range(1000):
      if i in list1:
        rank1[i - 1] = list1.index(i)

      if i in list2:
        rank2[i - 1] = list2.index(i)

    df = pd.DataFrame([rank1, rank2], index=['rank1', 'rank2'])
    df = df.fillna(np.nan)
    df = df.transpose()
    # normalize by replacing missing values
    df = rank_fillna(df)
    return df, rank1, rank2, dict_full



def rank_eval(rank_true, dfmerged, nrank):
  """
    Returns a dataframe where the columns are the rank aggregation algorithms
    and the rows are the metrics used to evaluate the algorithms

    Keyword arguments:
    rank_true: a list containing the accurate rank for each item
    dfmerged: a dataframe where each row represents an item and each column represents
              the rank aggregation algorithms
    nrank: the total number of elements to be ranked
  """
  columns = dfmerged.columns
  df = pd.DataFrame([], index=['spearman_rho', 'kendall-tau'])
  for col in columns:
    rank_list    = dfmerged[col].tolist()
    spearman_rho = scipy.stats.spearmanr(rank_true[0:nrank], rank_list[0:100])
    kendall_tau  = scipy.stats.kendalltau(rank_true[0:nrank], rank_list[0:100])
    df[col]      = [spearman_rho[0], kendall_tau[0]]
  return df






#####################################################################################
def rank_merge(df, method='borda'):
    """
      Returns a list of merged ranks for each item in the dataframe

      Keyword arguments:
      df: a dataframe in which each row represents an item and each column
          represents it's rank assigned by a ranker
      method: A rank aggregation algorithm
    """
    columns = df.columns
    # rank aggregation algorithm for dowdall and average doesn't accept dataframes
    # so convert each column to list
    rank_list = [df[column].tolist() for column in columns]
    if method == 'borda':
      rank = rk.rank(rk.borda(df)).tolist()

    elif method == 'majority':
      rank = rk.majority(df).tolist()

    elif method == 'center':
      rank = rk.center(df).tolist()

    elif method == 'instant_runoff':
      agg = ra.RankAggregator()   
      rank = agg.instant_runoff(rank_list)

    elif method == 'dowdall':
      agg = ra.RankAggregator()
      rank = agg.dowdall(rank_list)
      res = []
      for r in rank:
        res.append(r[0])
      return res

    elif method == 'average':
      agg = ra.RankAggregator()
      rank = agg.average_rank(rank_list)
      res = [None] * 1000
      for r in rank:
        res[int(r[0]) - 1] = r[1]
      return res

    elif method == 'mc4':
      aggregated_ranks = mc4_aggregator(df, header_row = 0, index_col = 0) 
      return list(aggregated_ranks.values())
    return rank


def rank_merge_v2(list1, list2, nrank):
  """ Inver MRR cores
    A custom rank aggregation algorithm that assigns a score to each item 
    based on its rankings. The algorithm sorts the items by their score in
    descending order and returns them

    Keyword arguments:
    list1: a list containing the ranks of items
    list2: a list containing the ranks of items
    nrank: total number of elements to be ranked
  """
  scores = []
  for i in range(1, nrank + 1):
    rank_1, rank_2 = nrank, nrank
    if i in list1:
      rank_1 = list1.index(i) + 1

    if i in list2:
      rank_2 = list2.index(i) + 1

    score = ((1 / rank_1) + (1 / rank_2)) / 2.0
    tup = (i, score)
    scores.append(tup)
  
  # sort list by rank in descending order
  scores.sort(key=lambda tup: tup[1], reverse=True)
  rank = []
  for (item, score) in scores:
    rank.append(item)
  return rank



